from django.db import connection, transaction

"""
PostgreSQL backend for Django DB Parti.

PostgreSQL partitioning is quit flexible, so if you can choose between database types,
PostgreSQL will be the best available choice for partitioned tables.
"""
class Postgresql(object):
    def __init__(self, table, partition_column, datetype):
        self.table = table
        self.cursor = connection.cursor()
        self.partition_column = partition_column
        self.datetype = datetype

    def partition_exists(self, partition_name):
        """Checks whether partition exists"""
        self.cursor.execute('SELECT EXISTS(SELECT 1 FROM information_schema.tables WHERE table_name=%s)',
            (self.table + partition_name,))
        return self.cursor.fetchone()[0]

    def create_partition(self, partition_name, fday, lday):
        """Creates partition with the given parameters"""
        self.cursor.execute('''
            -- We need to create a table first
            CREATE TABLE {child_table} (
                CHECK ({partition_column} >= {datetype} '{fday}' AND {partition_column} <= {datetype} '{lday}')
            ) INHERITS ({parent_table});

            -- Then we create an index to speed up things a little
            CREATE INDEX {child_table}_{partition_column} ON {child_table} ({partition_column});
        '''.format(
            child_table=self.table + partition_name,
            parent_table=self.table,
            partition_column=self.partition_column,
            fday=fday,
            lday=lday,
            datetype='DATE' if self.datetype == 'date' else 'TIMESTAMP',
        ))

        transaction.commit_unless_managed()

    def init_partition(self, partition_range):
        """Initializes needed triggers and functions for those triggers"""
        self.cursor.execute('''
            -- We need to create a before insert function
            CREATE OR REPLACE FUNCTION {parent_table}_insert_child()
            RETURNS TRIGGER AS $$
                DECLARE tablename TEXT;
                BEGIN
                    tablename := '{parent_table}_' || to_char(NEW.{partition_column}, '{partition_pattern}');
                    EXECUTE 'INSERT INTO ' || tablename || ' VALUES (($1).*);' USING NEW;
                    RETURN NEW;
                END;
            $$ LANGUAGE plpgsql;

            -- Then we create a trigger which calls the before insert function
            DO $$
            BEGIN
            IF NOT EXISTS(
                SELECT 1 
                FROM information_schema.triggers
                WHERE event_object_table = '{parent_table}'
                AND trigger_name = 'before_insert_{parent_table}_trigger'
            ) THEN
                CREATE TRIGGER before_insert_{parent_table}_trigger
                    BEFORE INSERT ON {parent_table}
                    FOR EACH ROW EXECUTE PROCEDURE {parent_table}_insert_child();
            END IF;
            END $$;

            -- Then we create a function to delete duplicate row from the master table after insert
            CREATE OR REPLACE FUNCTION {parent_table}_delete_master()
            RETURNS TRIGGER AS $$
                BEGIN
                    DELETE FROM ONLY {parent_table} WHERE id = NEW.id;
                    RETURN NEW;
                END;
            $$ LANGUAGE plpgsql;

            -- Lastly we create the after insert trigger that calls the after insert function
            DO $$
            BEGIN
            IF NOT EXISTS(
                SELECT 1 
                FROM information_schema.triggers
                WHERE event_object_table = '{parent_table}'
                AND trigger_name = 'after_insert_{parent_table}_trigger'
            ) THEN
                CREATE TRIGGER after_insert_{parent_table}_trigger
                    AFTER INSERT ON {parent_table}
                    FOR EACH ROW EXECUTE PROCEDURE {parent_table}_delete_master();
            END IF;
            END $$;
        '''.format(
            parent_table=self.table,
            partition_column=self.partition_column,
            partition_pattern=self.pattern_for_partition_range(partition_range),
        ))

        transaction.commit_unless_managed()

    @classmethod
    def pattern_for_partition_range(cls, partition_range):
        """Returns datetime pattern depending on the given partition range"""
        patterns = {
            'week': '"y"IYYY"w"IW',
            'month': '"y"YYYY"m"MM',
        }

        return patterns[partition_range]



from django.db import connection, transaction

"""
MySQL backend for Django DB Parti.

Thing to keep in mind is that MySQLdb doesn't support compound MySQL statements,
so each separate statement should be executed inside each self.cursor.execute().
Also MySQL has a lot of limitations for partitioned tables, please see MySQL site.
"""
class Mysql(object):
    def __init__(self, table, partition_column, datetype):
        self.table = table
        self.cursor = connection.cursor()
        self.partition_column = partition_column
        self.datetype = datetype

    def partition_exists(self, partition_name):
        """Checks whether partition exists"""
        self.cursor.execute('''
            SELECT EXISTS(
                SELECT 1 FROM information_schema.partitions
                WHERE table_name='{parent_table}' AND partition_name='{child_table}');
        '''.format(
            parent_table=self.table,
            child_table=partition_name.replace('_', ''),
        ))
        return self.cursor.fetchone()[0]

    def create_partition(self, partition_name, fday, lday):
        """Creates partition with the given parameters"""
        self.cursor.execute('''
            ALTER TABLE {parent_table} ADD PARTITION (
                PARTITION {child_table} VALUES LESS THAN ({datetype}('{lday}') + {addition})
            );
        '''.format(
            child_table=partition_name.replace('_', ''),
            parent_table=self.table,
            lday=lday,
            addition='1' if self.datetype == 'date' else '86400',
            datetype='TO_DAYS' if self.datetype == 'date' else 'UNIX_TIMESTAMP',
        ))

        transaction.commit_unless_managed()

    def init_partition(self, partition_range):
        """Initializes needed database stuff"""
        self.cursor.execute('''
            -- We need to rebuild primary key for our partitioning to work
            ALTER table {parent_table} DROP PRIMARY KEY, add PRIMARY KEY (`id`,`{partition_column}`);
        '''.format(
            parent_table=self.table,
            partition_column=self.partition_column,
        ))

        self.cursor.execute('''
            -- Then we create zero partition to speed up things due to the partitioning
            -- implementation in the early versions of MySQL database (see bug #49754)
            ALTER TABLE {parent_table} PARTITION BY RANGE ({datetype}({partition_column}))(
                PARTITION {partition_pattern} VALUES LESS THAN (0)
            );
        '''.format(
            parent_table=self.table,
            partition_column=self.partition_column,
            partition_pattern=self.pattern_for_partition_range(partition_range),
            datetype='TO_DAYS' if self.datetype == 'date' else 'UNIX_TIMESTAMP',
        ))

        transaction.commit_unless_managed()

    @classmethod
    def pattern_for_partition_range(cls, partition_range):
        """Returns datetime pattern depending on the given partition range"""
        patterns = {
            'week': 'y0000w00',
            'month': 'y0000m00',
        }

        return patterns[partition_range]



from django.contrib import admin
from dbparti.utilities import DateTimeMixin


class PartitionableAdmin(DateTimeMixin, admin.ModelAdmin):
    partition_show = 'all'

    def __init__(self, *args, **kwargs):
        """Initializes all the parent classes"""
        admin.ModelAdmin.__init__(self, *args, **kwargs)
        DateTimeMixin.__init__(
            self,
            self.partition_show,
            self.opts.partition_range,
            None,
            self.opts.get_field(self.opts.partition_column).get_internal_type(),
        )

    def queryset(self, request):
        """Determines data from what partitions should be shown in django admin"""
        fday, lday = self.get_partition_show_period(self.get_datetime_string('date'))
        qs = super(PartitionableAdmin, self).queryset(request)

        if fday is None and lday is None:
            qs = qs.all()
        else:
            qs = qs.filter(
                **{self.opts.partition_column + '__gte': fday, self.opts.partition_column + '__lte': lday}
            )

        return qs



from django.db import models, connection
from dbparti.utilities import DateTimeMixin


models.options.DEFAULT_NAMES += ('partition_range', 'partition_column')

class Partitionable(DateTimeMixin, models.Model):
    def __init__(self, *args, **kwargs):
        """Initializes all the parent classes and the current database backend"""
        models.Model.__init__(self, *args, **kwargs)
        DateTimeMixin.__init__(
            self,
            False,
            self._meta.partition_range,
            getattr(self, self._meta.partition_column),
            self._meta.get_field(self._meta.partition_column).get_internal_type(),
        )

        self.db = getattr(__import__('dbparti.backends.' + connection.vendor,
            fromlist=[connection.vendor.capitalize()]), connection.vendor.capitalize())(
                self._meta.db_table, self._meta.partition_column, self.get_datetype())

    def save(self, *args, **kwargs):
        """Determines into what partition the data should be saved"""
        fday, lday = self.get_partition_range_period(self.get_datetime_string('date'))

        if not self.db.partition_exists(self.get_partition_name()):
            self.db.create_partition(self.get_partition_name(), fday, lday)

        super(Partitionable, self).save(*args, **kwargs)

    class Meta:
        abstract = True
        partition_range = 'month'
        partition_column = 'partdate'



from datetime import datetime, timedelta


class DateTimeMixin(object):
    PARTITION_SHOWS = ('all', 'current', 'previous')
    PARTITION_RANGES = ('month', 'week')

    def __init__(self, partition_show, partition_range, partition_column_val, partition_column_type):
        """Initializes all the datetime logic needed for datetime partitioning"""
        self.partition_show = partition_show if partition_show else self.PARTITION_SHOWS[0]
        self.partition_range = partition_range if partition_range else self.PARTITION_RANGES[0]
        self.partition_column_val = partition_column_val
        self.partition_column_type = partition_column_type

        if self.partition_show not in self.PARTITION_SHOWS:
            raise Exception('Unknown partition show type "{0}", allowed types are: {1}'.format(
                self.partition_show, ', '.join(self.PARTITION_SHOWS)))
        if self.partition_range not in self.PARTITION_RANGES:
            raise Exception('Unknown partition range type "{0}", allowed types are: {1}'.format(
                self.partition_range, ', '.join(self.PARTITION_RANGES)))

        self.datestring = '%Y-%m-%d'
        self.timestring = '%Y-%m-%d %H:%M:%S.%f'

        if partition_column_type == 'DateField':
            self.datetype = 'date'
            self.workdate = datetime.strptime(str(self.partition_column_val), self.datestring).date() \
                if self.partition_column_val is not None else datetime.now().date()
        else:
            self.datetype = 'time'
            self.workdate = datetime.strptime(str(self.partition_column_val), self.timestring) \
                if self.partition_column_val is not None else datetime.now()

    def get_datetype(self):
        """Returns the current datetype (date or time)"""
        return self.datetype

    def get_datetime_string(self, dttype):
        """Returns datetime string compatible with strftime depending on datetype"""
        return self.datestring if dttype == 'date' else self.timestring

    def get_partition_name(self):
        """Returns partition name depending on the current workdate"""
        return '_' + self.workdate.strftime(getattr(self, 'partition_name_for_' + self.partition_range)())

    def get_partition_range_period(self, timestring):
        """Returns partition period start and end depending on the current workdate and datetype"""
        return getattr(self, 'partition_period_for_' + self.partition_range)(self.workdate, timestring)

    def get_partition_show_period(self, timestring):
        """Returns partition period start and end for the django admin"""
        if self.partition_show == 'all':
            return None, None
        elif self.partition_show == 'previous':
            if self.partition_range == 'week':
                dt = datetime.now().date() - timedelta(weeks=1)
            elif self.partition_range == 'month':
                today = datetime.now().date()
                dt = datetime(today.year, today.month, 1) - timedelta(days=1)
        elif self.partition_show == 'current':
            dt = datetime.now().date()

        return getattr(self, 'partition_period_for_' + self.partition_range)(dt, timestring)

    @classmethod
    def partition_name_for_week(cls):
        """Returns strftime compatible partition name part for week partition range"""
        return 'y%Yw%V'

    @classmethod
    def partition_name_for_month(cls):
        """Returns strftime compatible partition name part for month partition range"""
        return 'y%Ym%m'

    @classmethod
    def partition_period_for_week(cls, today, timestring):
        """Returns partition period start and end for week partition range"""
        dt = datetime(today.year, 1, 1)

        if dt.weekday() > 3:
            dt = dt + timedelta(7 - dt.weekday())
        else:
            dt = dt - timedelta(dt.weekday())

        dlt = timedelta(days=(int(today.strftime('%V')) - 1) * 7)
        fday = (dt + dlt).strftime(timestring)
        lday = (dt + dlt + timedelta(days=6)).strftime(timestring)

        return fday, lday

    @classmethod
    def partition_period_for_month(cls, today, timestring):
        """Returns partition period start and end for month partition range"""
        fday = datetime(today.year, today.month, 1).strftime(timestring)
        lday = (datetime(today.year, today.month + 1, 1) - timedelta(days=1)).strftime(timestring)
        return fday, lday
